
; This code implements SLEEP as well (it has to know about sleepers, so it
; might as well do both).

(define wakeup-service-lock (make-lock))

; Put the blocked thread on a list.

(define (handle-blocked-channel channel)
  (with-lock wakeup-service-lock
    (lambda ()
      (add-to-queue (current-thread)
		    unspecific
		    (channel-queue channel))))
  (enter-kernel!)
  (suspend-this-thread))

; An a-list of (<channel> . <queue>).

(define *blocked-i/o-threads* '())

(define (waiting-for-i/o?)
  (not (null? *blocked-i/o-threads*)))

(define (channel-queue channel)
  (cond ((assq channel *blocked-i/o-threads*)
	 => cdr)
	(else
	 (let ((q (make-queue)))
	   (set! *blocked-i/o-threads*
		 (cons (cons channel q)
		       *blocked-i/o-threads*))
	   q))))

; CHANNELS must be in the same order as *BLOCKED-I/O-THREADS*.

(define (wake-up-some-i/o-blocked-threads channels)   ; interrupts disabled
  (let loop ((channels channels) (a-list *blocked-i/o-threads*) (rem '()))
    (cond ((null? channels)
	   (set! *blocked-i/o-threads* (append a-list rem)))
	  ((eq? (car channels) (car (car a-list)))
	   (let ((q (cdr (car a-list))))
	     (do ()
		 ((queue-empty? q))
	       (make-cell-ready! (dequeue q))))
	   (loop (cdr channels) (cdr a-list) rem))
	  ((queue-empty? (cdar a-list))
	   (loop channels (cdr a-list) rem))
	  (else
	   (loop channels (cdr a-list) (cons (car a-list) rem))))))

; There is a race condition here if interrupts are not disabled: a channel
; could be in the list returned even though all threads waiting on it have
; been killed.

(define (blocked-channels)   ; interrupts disabled
  (let recur ((to-do *blocked-i/o-threads*))
    (cond ((null? to-do)
	   '())
	  ((queue-empty? (cdr (car to-do)))
	   (recur (cdr to-do)))
	  (else
	   (cons (car (car to-do))
		 (recur (cdr to-do)))))))


; Sleeping for N micro-seconds.

(define (sleep n)
  (obtain-lock wakeup-service-lock)
  (enter-kernel!)
  (let ((q (make-queue)))  ; will only have one entry, but it has to be a queue
    (add-to-queue (current-thread) unspecific q)
    (set! *dozers*
	  (insert (cons (+ (real-time) n) q)
		  *dozers*
		  (lambda (frob1 frob2)
		    (< (car frob1) (car frob2)))))
    (release-lock wakeup-service-lock)
    (suspend-this-thread)))

(define *dozers* '())  ;List of (wakeup-time . queue)

(define (wake-up-some-threads)    ; interrupts disabled
  (cond ((null? *dozers*)
	 #f)
	((queue-empty? (cdar *dozers*))
	 (set! *dozers* (cdr *dozers*))
	 (wake-up-some-threads))
	((< (real-time) (car (car *dozers*)))
	 #f)
	(else
	 (let ((queue (cdr (car *dozers*))))
	   (set! *dozers* (cdr *dozers*))
	   (if (not (queue-empty? queue))
	       (make-cell-ready! (dequeue queue)))
	   (wake-up-some-threads)))))

(define (insert x l <)
  (cond ((null? l) (list x))
	((< x (car l)) (cons x l))
	(else (cons (car l) (insert x (cdr l) <)))))


; Wakeup service

(define (initialize-sleepers)
  (set! wakeup-service-lock (make-lock)) ; someone may be on the old one
  (set! *dozers* '())
  (set! *blocked-i/o-threads* '()))
  
(define (wakeup-service)
  (dynamic-wind
   unspecific
   (lambda ()
     (enter-kernel!)
     (let loop ()
       (obtain-lock wakeup-service-lock)
       (if (not (null? *dozers*))
	   (wake-up-some-threads))
       (force-output-ports!)
       (let ((channels (blocked-channels)))
	 (cond ((not (queue-empty? runnable-threads))
		(check-for-i/o channels))
	       ((not (null? channels))
		(wait-for-i/o channels))
	       ((null? *dozers*)
		(signal 'deadlock)
		(error "no threads to run"))
	       (else
		(set! *blocked-i/o-threads* '())
		(wait-for-dozers))))
       (release-lock wakeup-service-lock)
       (relinquish-timeslice)
       (loop)))
   (lambda ()
     ;; If wakeup service gets killed, propagate kill to the threads
     ;; it was going to wake up, so their unwind forms can run.
     (for-each (lambda (dozer)
		 (kill-thread (cell-thread (cdr dozer))))
	       *dozers*)
     (for-each (lambda (p)
		 (for-each (lambda (cell)
			     (if (cell-thread cell)
				 (kill-thread (cell-thread cell))))
			   (cdr p)))
	       *blocked-i/o-threads*)
     (initialize-sleepers)
     (if (eq? (lock-owner wakeup-service-lock) (current-thread))
	 (release-lock wakeup-service-lock)))))

; If there is nothing else to do, wait either forever or until the next
; sleeper is due to wake.

(define find-ready-channels (structure-ref channels find-ready-channels))

(define (wait-for-i/o channels)
  (maybe-wait-for-i/o channels #t))

(define (check-for-i/o channels)
  (maybe-wait-for-i/o channels #f))

(define (wait-for-dozers)
  (maybe-wait-for-i/o '() #t))

(define (maybe-wait-for-i/o channels wait?)	; interrupts disabled
  (stop-periodic-interrupts!)
  (let* ((now (real-time))
	 (time (if wait?
		   (if (null? *dozers*)
		       #f   ; forever
		       (max 0 (- (car (car *dozers*)) now)))
		   0))
	 (channels (find-ready-channels time -3 channels)))
    (start-periodic-interrupts!)
    (if (not (null? channels))
	(wake-up-some-i/o-blocked-threads channels))))	

(define (read-char-with-timeout port t)
  (let* ((waiting? #t)
	 (cv (make-condvar))
	 (reader (spawn (lambda ()
			  (let ((char (read-char port)))
			    ;; depends on no interrupts here
			    (set! waiting? #f)
			    (condvar-set! cv char)))))
	 (sleeper (spawn (lambda ()
			   (with-interrupts-inhibited
			    (lambda ()
			      (sleep t)
			      (cond (waiting?
				     (kill-thread reader)
				     (condvar-set! cv #f))))))))
	 (res (condvar-ref cv)))
    (kill-thread sleeper)
    res))
