/*Copyright (c) 1994 by Richard Kelsey and Jonathan Rees.  See file COPYING.*/


/* Implementation of the vm-extension opcode.  This is completely
   optional; nothing in the standard system uses these features.
   If you have ANSI C but not POSIX support, try compiling with -DPOSIX=0.

   fdopen: POSIX.1
   getenv: POSIX.1, ANSI C
   setuid, setgid: POSIX.1
   popen: POSIX.2
   floating point: POSIX.1, ANSI C (should we be linking with -lM or -lm?)
   sprintf: POSIX.1, ANSI C
   atof: POSIX.1, ANSI C
   chroot: not standard

 */

#ifndef POSIX
#  define POSIX 2
#endif

#include <stdio.h>
#include "sysdep.h"
#include "scheme48.h"

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <signal.h>
#include <unistd.h>		/* setuid & setgid */

/* #include <fcntl.h>	*/	/* for O_RDWR */

#define GREATEST_FIXNUM_VALUE ((1 << 29) - 1)
#define LEAST_FIXNUM_VALUE (-1 << 29)
#define CHANNEL_INDEX(x) EXTRACT_FIXNUM(STOB_REF(x, 1))
#define FOR_INPUT 1
#define FOR_OUTPUT 2

int channel_to_fd(scheme_value);

typedef struct {
  char b[sizeof(double)];
} unaligned_double;

typedef union {
  double f;
  unaligned_double b;
} float_or_bytes;

extern long Sextension_valueS;  /* how values are returned */

/* return status values */
#define EXT_ST_OKAY 0
#define EXT_ST_EXCEPTION 1

#define EXT_RETURN(value) {Sextension_valueS = (value); return EXT_ST_OKAY; }
#define EXT_EXCEPTION     return EXT_ST_EXCEPTION

/******************************************/

scheme_value
extended_vm (long key, scheme_value value)
{
  double x, y;

  switch (key) {

    /* Cases 0 through 19 are reserved for the mobot system. */

  case 0:			/* read jumpers on 68000 board */
     EXT_RETURN(ENTER_FIXNUM(0));

#define SOCKET_SUPPORT

#if defined(SOCKET_SUPPORT)
  case 20:
    { extern int internet_stream_socket();
      int s = internet_stream_socket();
      if (s < 0)
	EXT_EXCEPTION;
      else
	EXT_RETURN(ENTER_FIXNUM(s));
    }

  case 21:
    { extern int socket_bind(int, int);
      int port = socket_bind(EXTRACT_FIXNUM(value), 0);
      if (port < 0)
	EXT_EXCEPTION;
      else
	EXT_RETURN(ENTER_FIXNUM(port));
    }

  case 22:
    { extern int socket_accept(int);
      int fd = socket_accept(channel_to_fd(value));
      if (fd > 0)
	EXT_RETURN(ENTER_FIXNUM(fd))
      else if (fd == -2)
	EXT_RETURN(ENTER_FIXNUM(-1))
      else
	EXT_EXCEPTION;
    }

  case 23:
    if (!PAIRP(value)) EXT_EXCEPTION;
    { extern int socket_connect(char *, int);
      char *hostname = &STRING_REF(CAR(value), 0);
      long port = EXTRACT_FIXNUM(CDR(value));
      int fd;
      fd = socket_connect(hostname, port);
      if (fd < 0)
	EXT_EXCEPTION;
      else
	EXT_RETURN(ENTER_FIXNUM(fd));
    }

#endif

  /* getenv() */
  case 26: {
    scheme_value env_var, result_buffer;
    char *result;
    size_t result_len;

    if (!PAIRP(value)) EXT_EXCEPTION;
    env_var = CAR(value);
    result_buffer = CDR(value);
    if (!STRINGP(env_var) || !STRINGP(result_buffer)) EXT_EXCEPTION;
    result = getenv(&STRING_REF(env_var, 0));
    if (result == NULL)
      EXT_RETURN(SCHFALSE);
    result_len = strlen(result);
    if (result_len > STRING_LENGTH(result_buffer))
      EXT_EXCEPTION;
    strncpy(&STRING_REF(result_buffer, 0), result, result_len);
    EXT_RETURN(ENTER_FIXNUM(result_len));
  }

#if POSIX
  case 27: {
    /* This is intended for use by HTTP scripts... */
    if (!PAIRP(value) || !FIXNUMP(CAR(value)) || !FIXNUMP(CDR(value)))
      EXT_EXCEPTION;
    if (setgid(EXTRACT_FIXNUM(CDR(value))) != 0) {
      perror("setgid");
      EXT_RETURN(SCHFALSE); }
    if (setuid(EXTRACT_FIXNUM(CAR(value))) != 0) {
      perror("setuid");
      EXT_RETURN(SCHFALSE); }
    else
      EXT_RETURN(SCHTRUE);
  }
#endif

#if defined(HAVE_CHROOT)
  case 28: {
    if (!STRINGP(value))
      EXT_EXCEPTION;
    else if (chroot(&STRING_REF(value, 0)) != 0) {
      perror("chroot");
      EXT_RETURN(SCHFALSE); }
    else
      EXT_RETURN(SCHTRUE);
  }
#endif

  /* dup() support */
  case 29:
    { long fd = dup(EXTRACT_FIXNUM(value));
      if (fd < 0) {
	EXT_EXCEPTION;
      }
      else
	EXT_RETURN(ENTER_FIXNUM(fd));
    }

#if POSIX >= 2
  case 96: {
    int status;
    if (!STRINGP(value))
      EXT_EXCEPTION;
    status = system(&STRING_REF(value, 0));
    if (status == -1) {
      perror("chroot");
      EXT_EXCEPTION; }
    else
      EXT_RETURN(ENTER_FIXNUM(status)); /* cf. waitpid() */
  }

  /* popen() support.  Rather kludgey; there's no pclose(), so
     zombies will pile up. */
/* This needs to be changed now that we have fd's instead of FILE *'s.
  case 97:
  case 98: {
    if (!PAIRP(value)) EXT_EXCEPTION;
    { scheme_value port = CAR(value);
      FILE **pstream = port_to_stream(port);
      scheme_value command = CDR(value);
      FILE *new_stream;
      struct sigaction action;

      if (pstream == NULL) EXT_EXCEPTION;

      action.sa_handler = SIG_IGN;
      action.sa_flags = 0;
      sigemptyset(&action.sa_mask);
      sigaction(SIGPIPE, &action, NULL);

      new_stream = popen(&STRING_REF(command, 0), key == 97 ? "r" : "w");
      if (new_stream == NULL) EXT_EXCEPTION;
      fclose(*pstream);
      *pstream = new_stream;
      EXT_RETURN(SCHTRUE);
    }
  }
*/
#endif /* POSIX.2 */


  /* Floating point */

#define FLOP 100
#define FLOP2(i) case FLOP+(i): \
		   if (!STOBP(value) || STOB_LLENGTH(value) != 2) \
		     EXT_EXCEPTION;
#define FLOP3(i) case FLOP+(i): \
		   if (!STOBP(value) || STOB_LLENGTH(value) != 3) \
		     EXT_EXCEPTION;

#define get_arg(args,i) STOB_REF(args,(i))
#define get_string_arg(args,i) (&STRING_REF(get_arg(args,i), 0))

#define get_float_arg(args, i, var) EXTRACT_FLOAT(get_arg(args, i), var)
#define set_float_arg(args, i, val) SET_FLOAT(get_arg(args, i), val)

#define EXTRACT_FLOAT(stob, var) \
  { scheme_value temp_ = (stob); \
    float_or_bytes loser_; \
    if (!STOBP(temp_)) EXT_EXCEPTION; \
    loser_.b = *(unaligned_double*)(&STOB_REF(temp_, 0)); \
    (var) = loser_.f; }

#define SET_FLOAT(stob, val) \
  { scheme_value temp_ = (stob); \
    float_or_bytes loser_; \
    if (!STOBP(temp_)) EXT_EXCEPTION; \
    loser_.f = (double)(val); \
    *(unaligned_double*)(&STOB_REF(temp_, 0)) = loser_.b; }

  FLOP3(0) {
    get_float_arg(value, 0, x);
    get_float_arg(value, 1, y);
    set_float_arg(value, 2, x + y);
    EXT_RETURN(UNSPECIFIC);}
  FLOP3(1) {
    get_float_arg(value, 0, x);
    get_float_arg(value, 1, y);
    set_float_arg(value, 2, x - y);
    EXT_RETURN(UNSPECIFIC);}
  FLOP3(2) {
    get_float_arg(value, 0, x);
    get_float_arg(value, 1, y);
    set_float_arg(value, 2, x * y);
    EXT_RETURN(UNSPECIFIC);}
  FLOP3(3) {
    get_float_arg(value, 0, x);
    get_float_arg(value, 1, y);
    if (y == 0.0) EXT_EXCEPTION;
    set_float_arg(value, 2, x / y);
    EXT_RETURN(UNSPECIFIC);}
  FLOP2(4) {
    get_float_arg(value, 0, x);
    get_float_arg(value, 1, y);
    EXT_RETURN(ENTER_BOOLEAN(x == y));}
  FLOP2(5) {
    get_float_arg(value, 0, x);
    get_float_arg(value, 1, y);
    EXT_RETURN(ENTER_BOOLEAN(x < y));}
  FLOP2(6) {			/* fixnum->float */
    scheme_value arg = get_arg(value, 0);
    if (!FIXNUMP(arg)) EXT_RETURN(SCHFALSE);
    set_float_arg(value, 1, EXTRACT_FIXNUM(arg));
    EXT_RETURN(SCHTRUE);}
  FLOP2(7) {			/* string->float */
    char *str = get_string_arg(value, 0);
    set_float_arg(value, 1, atof(str));
    EXT_RETURN(UNSPECIFIC);}
  FLOP2(8) {			/* float->string */
    size_t len;
    char *str = get_string_arg(value,1);
    get_float_arg(value, 0, x);
    sprintf(str, "%g", x);
    len = strlen(str);
    if (len > STRING_LENGTH(get_arg(value,1)))
      /* unlikely but catastrophic */
      fprintf(stderr, "printing float: output too long: %s\n",
	      str);
    EXT_RETURN(ENTER_FIXNUM(len));}

    /* exp log sin cos tan asin acos atan sqrt */

  FLOP2(9) {
    get_float_arg(value, 0, x);
    set_float_arg(value, 1, exp(x));
    EXT_RETURN(UNSPECIFIC);}
  FLOP2(10) {
    get_float_arg(value, 0, x);
    set_float_arg(value, 1, log(x));
    EXT_RETURN(UNSPECIFIC);}
  FLOP2(11) {
    get_float_arg(value, 0, x);
    set_float_arg(value, 1, sin(x));
    EXT_RETURN(UNSPECIFIC);}
  FLOP2(12) {
    get_float_arg(value, 0, x);
    set_float_arg(value, 1, cos(x));
    EXT_RETURN(UNSPECIFIC);}
  FLOP2(13) {
    get_float_arg(value, 0, x);
    set_float_arg(value, 1, tan(x));
    EXT_RETURN(UNSPECIFIC);}
  FLOP2(14) {
    get_float_arg(value, 0, x);
    set_float_arg(value, 1, asin(x));
    EXT_RETURN(UNSPECIFIC);}
  FLOP2(15) {
    get_float_arg(value, 0, x);
    set_float_arg(value, 1, acos(x));
    EXT_RETURN(UNSPECIFIC);}
  FLOP3(16) {			/* atan */
    get_float_arg(value, 0, y);
    get_float_arg(value, 1, x);
    set_float_arg(value, 2, atan2(y, x));
    EXT_RETURN(UNSPECIFIC);}
  FLOP2(17) {
    get_float_arg(value, 0, x);
    set_float_arg(value, 1, sqrt(x));
    EXT_RETURN(UNSPECIFIC);}

  FLOP2(18) {			/* floor */
    get_float_arg(value, 0, x);
    set_float_arg(value, 1, floor(x));
    EXT_RETURN(UNSPECIFIC);}
  case FLOP+19: {		/* integer? */
    EXTRACT_FLOAT(value, x);
    EXT_RETURN(ENTER_BOOLEAN(fmod(x, 1.0) == 0.0)); }
  case FLOP+20: {		/* float->fixnum */
    EXTRACT_FLOAT(value, x);
    if (x <= (double)GREATEST_FIXNUM_VALUE
	&& x >= (double)LEAST_FIXNUM_VALUE)
      {
      EXT_RETURN(ENTER_FIXNUM((long)x)); }
    else
      EXT_RETURN(SCHFALSE);}
  FLOP3(21) {			/* quotient */
    double z;
    get_float_arg(value, 0, x);
    get_float_arg(value, 1, y);
    if (fmod(x, 1.0) != 0.0 || fmod(y, 1.0) != 0.0) EXT_EXCEPTION;
    if (y == 0.0) EXT_EXCEPTION;
    z = x / y;
    set_float_arg(value, 2, z < 0.0 ? ceil(z) : floor(z));
    EXT_RETURN(UNSPECIFIC);}
  FLOP3(22) {			/* remainder */
    get_float_arg(value, 0, x);
    get_float_arg(value, 1, y);
    if (fmod(x, 1.0) != 0.0 || fmod(y, 1.0) != 0.0) EXT_EXCEPTION;
    if (y == 0.0) EXT_EXCEPTION;

    /* "fmod(double x, double y) returns the floating-point remainder
       (f) of the division of x by y, where f has the same sign as x,
       such that x=iy+f for some integer i, and |f| < |y|." */

    set_float_arg(value, 2, fmod(x, y));
    EXT_RETURN(UNSPECIFIC);}

  default:
    EXT_EXCEPTION;
  }
}


int
channel_to_fd(scheme_value channel)
{
  long index;
  extern int *Sopen_portsS;

  if (!CHANNELP(channel))
    return -1;		/* not a channel */

  index = CHANNEL_INDEX(channel);
  if (index < 0)
    return -1;		/* channel not open */

  return Sopen_portsS[index];
}


#if defined(SOCKET_SUPPORT)

/*
 * Socket support for Scheme 48.  Note, the division of labor is rather
 * silly.  Given that socket_connect() creates it's own socket, there is
 * no reason for socket_bind() to not do the same (and to eliminate
 * internet_stream_socket()).  The interface to extension() is a little
 * easier this way, because it never has to return two values.
 * This code was written by a loyal member of the state.  As such, I of
 * course ask nothing in return.  From each according to their ability,
 * to each according to their need.  I have what I need.
 * I don't need a copyright.
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <assert.h>


#define	LISTENQ		5

/*
 * Create an internet-domain stream (reliable, sequenced) socket.
 * We return the socket (a small non-negative integer) on success
 * or a negative integer on failure.
 */
int
internet_stream_socket(void)
{
  return (socket(AF_INET, SOCK_STREAM, 0));
}

/*
 * Given an internet-domain stream socket and a port number, bind
 * the socket to the port and prepare to receive connections.  If
 * the port number is 0 then we bind the socket to any available
 * port.  If all goes well, we return the port number we are bound
 * to.  It not, we return -1.
 */
int
socket_bind(int sock, int port)
{
  struct sockaddr_in	addr;
  int			len;
  
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(port);
  if ((bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0)
      ||  (listen(sock, LISTENQ) < 0))
    return (-1);
  if (port == 0) {
    len = sizeof(addr);
    if ((getsockname(sock, (struct sockaddr *)&addr, &len) < 0)
	||  (addr.sin_family != AF_INET))
      return (-1);
    port = htons(addr.sin_port);
  }
  return (port);
}

/*
 * Given an internet-domain stream socket which has been bound (via
 * socket_bind), accept a connection and return the resulting socket
 * (a small, non-negative integer).  If this fails, return a negative
 * integer.  select() is used to see if the accept() will block, in
 * which case -2 is returned.
 */
int
socket_accept(int sock)
{
  int			len;
  int			ready;
  struct sockaddr_in	addr;

#if defined(HAVE_SELECT)
  fd_set readfds;
  struct timeval timeout;

  FD_ZERO(&readfds);
  FD_SET(sock, &readfds);
  timerclear(&timeout);

  /* Under HPUX, select() is declared
      extern int select(size_t, int *, int *, int *, const struct timeval *);
     in sys/time.h.  This is probably Posix, but there's no knowing.
     Under SunOS, the int *'s are fd_set *'s. */
  ready = select(FD_SETSIZE,
#if defined(hpux)
		(int *)
#endif
		&readfds,
		NULL,
#if defined(hpux)
		(int *)
#endif
		&readfds, &timeout);
  if (ready == -1) {
    return -1; }
  else if (ready == 0) {
    return -2; }
#endif

  len = sizeof(addr);
  return (accept(sock, (struct sockaddr *)&addr, &len));
}

/*
 * Given a machine name and a port number, connect the socket to that
 * machine/port.  If the machine name is the empty string, use the local
 * machine.  If this succeeds return a small non-negative integer.
 * If it fails, return a negative integer.
 */
int
socket_connect(char *mach, int port)
{
  char			mbuff[MAXHOSTNAMELEN];
  struct hostent	*hp;
  struct sockaddr_in	addr;
  int			sock;
  
  if (*mach == '\0') {
    if (gethostname(mbuff, sizeof(mbuff)) < 0)
      return (-1);
    mach = mbuff;
  }
  hp = gethostbyname(mach);
  if (hp == NULL)
    return (-1);
  bzero((char *)&addr, sizeof(addr));
  addr.sin_family = hp->h_addrtype;
  if (hp->h_length > sizeof(addr.sin_addr))
    return (-1);
  bcopy(hp->h_addr, (char *)&addr.sin_addr, hp->h_length);
  addr.sin_port = htons(port);
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0)
    return (-1);
  if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
    close(sock);
    return (-1);
  }
  return (sock);
}

#endif

