
; Simple buffered I/O for reading and writing heap images.

(define buffer-size 1024)

; Output

(define *current-output-port* (current-output-port))

(define *output-buffer* #f)
(define *output-index* 0)
(define *output-buffer-limit* 0)
(define *output-errors?* #t)

(define (write-output-to-port port)
  (let ((buffer (allocate-memory buffer-size)))
    (cond ((not (null-pointer? buffer))
	   (set! *current-output-port* port)
	   (set! *output-buffer* buffer)
	   (set! *output-index* buffer)
	   (set! *output-limit* (+ buffer buffer-size))
	   (set! *output-errors?* #f)
	   #t)
	  (else
	   (set! *output-errors?* #t)
	   #f))))

(define (finish-buffered-output)
  (flush-output-buffer)
  (let ((errors? *output-errors?*))
    (deallocate-memory *output-buffer*)
    (set! *output-errors?* #t)
    (not errors?)))

(define (flush-output-buffer)
  (cond (*output-errors?*)
	((really-write-bytes *output-buffer* (- *output-index* *output-buffer*))
	 (set! *output-index* *output-buffer*))
	(else
	 (set! *output-errors?* #t)))
  *output-errors?)

(define (write-bytes-to-buffer start count)
  (if (not (and (flush-output-buffer)
		(really-write-bytes start count)))
      (set! *output-errors?* #t)))

(define (really-write-bytes start count)
  (let loop ((start start) (count count))
    (if (= count 0)
	#t
	(write-block *current-output-port*
		     start
		     count
		     (lambda (okay? sent)
		       (if okay?
			   (loop (+ start sent) (- count sent))
			   #f))))))

(define (write-string-to-buffer string)
  (if (not *output-errors?*)
      (let ((len (string-length string))
	    (limit *output-buffer-limit*))
	(let loop ((from 0) (to *output-index*))
	  (cond ((= from len)
		 (set! *output-index* to))
		((< to *output-buffer-limit*)
		 (unsigned-byte-set! to (string-ref string from))
		 (loop (+ from 1) (+ to 1)))
		((flush-output-buffer)
		 (loop from *output-buffer*)))))))

(define (write-char-to-buffer char)
  (cond ((and (not *output-errors?*)
	      (or (< *output-index* *output-buffer-limit*)
		  (flush-output-buffer)))
	 (unsigned-byte-set! *output-index* (char->ascii char))
	 (set! *output-index* (+ 1 *output-index*)))))

(define (buffer-newline)
  (write-char-to-buffer #\newline))

(define (write-number x)
  (if (not *output-errors*)
      (begin
	(really-write-number (cond ((< x 0)
				    (write-char #\-)
				    (- 0 x))
				   (else
				    x)))
	(buffer-newline))))
  
(define (really-write-number x)
  (and (if (>= x 10)
	   (really-write-number (quotient x 10))
	   #f)
       (write-char-to-buffer (ascii->char (+ (remainder x 10)
					     (char->ascii #\0))))))

;----------------------------------------------------------------
; Input

(define *current-input-port* (current-input-port))

(define *input-buffer* #f)
(define *input-index* 0)
(define *input-buffer-limit* 0)
(define *input-errors?* #t)
(define *input-eof?* #t)

(define (read-input-from-port port)
  (let ((buffer (allocate-memory buffer-size)))
    (cond ((not (null-pointer? buffer))
	   (set! *current-input-port* port)
	   (set! *input-buffer* buffer)
	   (set! *input-index* buffer)
	   (set! *input-limit* (+ buffer buffer-size))
	   (set! *input-errors?* #f)
	   (set! *input-eof?* #f)
	   #t)
	  (else
	   (set! *input-errors?* #t)
	   #f))))

(define (finish-buffered-input)
  (let ((errors? *input-errors?*))
    (deallocate-memory *input-buffer*)
    (set! *input-errors?* #t)
    (not errors?)))

(define (input-error?)
  *input-error?*)

(define (input-eof?)
  *input-eof?*)

(define (fill-buffer)
  (read-block *input-buffer* buffer-size
	      (lambda (okay? eof? got)
		(cond ((not okay?)
		       (set! *input-errors?* #t)
		       #f)
		      (eof?
		       (set! *input-errors?* #t)
		       (set! *input-eof?* #t)
		       #f)
		      ((= got 0)
		       (fill-buffer))
		      (else
		       (set! *input-index* *input-buffer*)
		       (set! *input-limit* (+ *input-buffer* got))
		       #t)))))

(define (skip-past-char char)
  (let ((target (char->ascii char)))
    (if (not *input-errors?*)
	(let loop ((at *input-index*) (end *input-limit*))
	  (cond ((= at end)
		 (if (fill-buffer)
		     (loop *input-index* *input-limit*)))
		((not (= char (unsigned-byte-ref at)))
		 (loop (+ at 1) end))
		(else
		 (set! *input-index* at)))))))

(define (read-bytes-from-buffer start count)
  (if (not *input-errors?*)
      (let ((used (min count (- *input-limit* *input-index*))
	(copy-bytes *input-index* start used)
	(set! *input-index* (+ *input-index* used))
	(if (not (= used count))
	    (let loop ((at (+ start used)) (count (- count used)))
	      (if (and (not *input-errors?*)
		       (> count 0))
		  (read-block at count
			      (lambda (okay? eof? got)
				(cond ((not okay?)
				       (set! *input-errors?* #t))
				      (eof?
				       (set! *input-errors?* #t)
				       (set! *input-eof?* #t))
				      (else
				       (loop (+ at got) (- count got))))))))))))))

(define (copy-bytes from to count)
  (let ((done (+ from count)))
    (do ((from from (+ from 1))
	 (to to (+ to 1)))
	((= from done))
      (unsigned-byte-set! to (unsigned-byte-ref from)))))

(define (read-number)
  (if *input-errors?*
      -1
      (let ((ch (let loop ()
		  (let ((ch (get-char)))
		    (if (digit? ch) ch (loop))))))
	(let loop ((r 0) (ch ch))
	  (if *input-errors?*
	      -1
	      (let ((r (+ (- (char->ascii ch) (char->ascii #\0))
			  (* r 10)))
		    (ch (get-char)))
		(if (digit? ch)
		    (loop r ch)
		    r)))))))

(define (get-char)
  (let ((at *input-index*))
    (cond ((< at *input-limit*)
	   (set! *input-index* (+ 1 at))
	   (ascii->char (unsigned-byte-ref at)))
	  ((fill-buffer)
	   (get-char))
	  (else
	   #\0))))  ; READ-NUMBER needs this to be a digit