
; Simple unbuffered I/O for reading and writing heap images.

; A one character buffer
(define *buffer* (unassigned))

(define (begin-simple-i/o)
  (set! *buffer* (allocate-memory 1))
  (cond ((<= 0 *buffer*)
	 (set! *output-error?* #f)
	 (set! *input-error?* #f)
	 (set! *input-eof?* #f)
	 #t)
	(else #f)))

(define (finish-simple-i/o)
  (deallocate-memory *buffer*)
  (let ((error? (or *output-error?*
		    *input-error?*)))
    (set! *output-error?* #t)   ; prevent accidental use of returned buffer
    (set! *input-error?* #t)
    (not error?)))

(define *output-error?* #f)

(define (maybe-write-string string port)
  (let ((len (string-length string)))
    (do ((i 0 (+ i 1)))
	((or (>= i len)
	     *output-error?*))
      (unsigned-byte-set! *buffer* (char->ascii (string-ref string i)))
      (let loop ()
	(write-block port *buffer* 1 (lambda (okay? sent)
				       (cond ((not okay?)
					      (set! *output-error?* #t))
					     ((= sent 0)
					      (loop)))))))))

(define (maybe-write-char char port)
  (cond ((not *output-error?*)
	 (unsigned-byte-set! *buffer* (char->ascii char))
	 (write-block port *buffer* 1 (lambda (okay? sent)
					(set! *output-error?* (not okay?)))))))

(define (maybe-write-number x port)
  (cond ((not *output-error?*)
	 (really-write-number (cond ((< x 0)
				     (maybe-write-char #\- port)
				     (- 0 x))
				    (else
				     x))
			      port)
	 (maybe-write-char #\newline port))))
  
(define (really-write-number x port)
  (if (>= x 10)
      (really-write-number (quotient x 10) port))
  (if (not *output-error?*)
      (maybe-write-char (ascii->char (+ (remainder x 10)
					(char->ascii #\0)))
			port)))

(define (maybe-write-block start count port)
  (if (not *output-error?*)
      (let loop ((start start) (count count))
	(if (> count 0)
	    (write-block port start count
			 (lambda (okay? sent)
			   (if okay?
			       (loop (+ start sent) (- count sent))
			       (set! *output-error?* #t))))))))

;----------------------------------------------------------------
; Input

(define *input-error?* #t)
(define *input-eof?* #t)
(define *peeked?* #f)

(define (input-error?)
  (and *input-error?*
       (not *input-eof?*)))

(define (input-eof?)
  *input-eof?*)

(define (maybe-read-char port)
  (cond (*peeked?*
	 (set! *peeked?* #f)
	 (ascii->char (unsigned-byte-ref *buffer*)))
	(*input-error?*
	 #\0)             ; a digit for MAYBE-READ-NUMBER
	(else
	 (read-block port *buffer* 1
		     (lambda (okay? eof? got)
		       (cond ((not okay?)
			      (set! *input-error?* #t)
			      #\0)             ; a digit for MAYBE-READ-NUMBER
			     (eof?
			      (set! *input-error?* #t)
			      (set! *input-eof?* #t)
			      #\0)             ; a digit for MAYBE-READ-NUMBER
			     ((= got 0)
			      (goto maybe-read-char port))
			     (else
			      (ascii->char (unsigned-byte-ref *buffer*)))))))))

(define (maybe-peek-char port)
  (let ((char (maybe-read-char port)))
    (set! *peeked?* #t)
    char))

(define (skip-past-char char port)
  (let loop ()
    (if (and (char=? char (maybe-read-char port))
	     (not *input-error?*))
	(loop))))

(define (maybe-read-number port)
  (if *input-error?*
      -1
      (let ((ch (let loop ()
		  (let ((ch (maybe-read-char port)))
		    (if (digit? ch) ch (loop))))))
	(let loop ((r 0) (ch ch))
	  (if *input-error?*
	      -1
	      (let ((r (+ (- (char->ascii ch) (char->ascii #\0))
			  (* r 10)))
		    (ch (maybe-peek-char port)))
		(if (digit? ch)
		    (loop r (maybe-read-char port))
		    r)))))))

(define (digit? char)
  (let ((i (char->ascii char)))
    (and (>= i (char->ascii #\0))
	 (<= i (char->ascii #\9)))))

(define (maybe-read-block start count port)
  (if (not *input-error?*)
      (let loop ((start start) (count count))
	(if (> count 0)
	    (read-block port start count
			(lambda (okay? eof? got)
			  (if (and okay? (not eof?))
			      (loop (+ start got) (- count got))
			      (set! *input-error?* #t))))))))
